"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const mongodb_1 = require("mongodb");
const projection_1 = require("./core/projection");
const tyr_1 = require("./tyr");
const path_1 = require("./core/path");
exports.metaRegex = /[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g;
function setFalse(v) {
    return v !== undefined && !v;
}
exports.setFalse = setFalse;
function escapeRegex(str) {
    return str.replace(exports.metaRegex, '\\$&');
}
exports.escapeRegex = escapeRegex;
function pathAdd(path, add) {
    return path ? path + '.' + add : add;
}
exports.pathAdd = pathAdd;
function isArrow(fn) {
    return /^[a-zA-Z0-9,_\s]*=>/.test(fn.toString());
}
exports.isArrow = isArrow;
function hasMongoUpdateOperator(update) {
    for (const key in update) {
        if (key.startsWith('$') && update.hasOwnProperty(key)) {
            return true;
        }
    }
    //return undefined;
}
exports.hasMongoUpdateOperator = hasMongoUpdateOperator;
function extractUpdateFields(doc, opts) {
    const updateFields = {};
    const projection = projection_1.extractProjection(opts);
    if (projection) {
        _.each(projection, (field, key) => {
            if (field) {
                updateFields[key] = 1;
            }
        });
    }
    else {
        _.each(doc, (field, key) => {
            if (key !== '_id') {
                updateFields[key] = 1;
            }
        });
    }
    return updateFields;
}
exports.extractUpdateFields = extractUpdateFields;
// Options parsing
// ===============
function isOptions(opts) {
    // TODO:  this is a hack, need to figure out a better way (though most likely a non-issue in practice)
    return (opts &&
        (opts.asOf !== undefined ||
            opts.auth !== undefined ||
            opts.author !== undefined ||
            opts.comment !== undefined ||
            opts.fields !== undefined ||
            opts.historical !== undefined ||
            opts.limit !== undefined ||
            opts.multi !== undefined ||
            opts.parallel !== undefined ||
            opts.perm !== undefined ||
            opts.populate !== undefined ||
            opts.projection !== undefined ||
            opts.query !== undefined ||
            opts.skip !== undefined ||
            opts.timestamps !== undefined ||
            opts.tyranid !== undefined ||
            opts.upsert !== undefined ||
            opts.writeConcern !== undefined ||
            !_.keys(opts).length));
}
exports.isOptions = isOptions;
function processOptions(collection, opts) {
    if (opts) {
        const fields = projection_1.extractProjection(opts);
        if (fields) {
            const f = projection_1.resolveProjection(collection.def.projections, fields);
            if (f !== fields) {
                const newOpts = {
                    ...opts,
                    projection: f,
                };
                delete newOpts.fields;
                return newOpts;
            }
        }
        return opts;
    }
    else {
        return {};
    }
}
exports.processOptions = processOptions;
function extractOptions(collection, args) {
    if (args.length && isOptions(args[args.length - 1])) {
        return processOptions(collection, args.pop());
    }
    else {
        return {};
    }
}
exports.extractOptions = extractOptions;
/**
 * This functions tries to avoid creating a new option object unless it has to.
 */
function combineOptions(...sources) {
    let first;
    let composite;
    for (const source of sources) {
        if (source) {
            if (composite) {
                _.assign(composite, source);
            }
            else if (first) {
                composite = {};
                _.assign(composite, first);
                _.assign(composite, source);
            }
            else {
                first = source;
            }
        }
    }
    return composite || first;
}
exports.combineOptions = combineOptions;
/**
 * Extracts the authorization out of a mongodb options-style object.
 */
function extractAuthorization(opts) {
    if (!opts) {
        return undefined;
    }
    const auth = opts.auth;
    if (auth) {
        //delete opts.auth;
        return auth === true ? tyr_1.default.local.user : auth;
    }
    // @deprecated
    const tyrOpts = opts.tyranid;
    if (tyrOpts) {
        delete opts.tyranid;
        if (tyrOpts.secure) {
            return tyrOpts.subject || tyrOpts.user || tyr_1.default.local.user;
        }
    }
    //return undefined;
}
exports.extractAuthorization = extractAuthorization;
/**
 * TODO: parseSaveObj() creates a new object and updates it with values to ensure that excess
 *       stuff on the object does not get written to the database which is good.
 *
 *       However, this means that any computed values/timestamp updates made here do not get made
 *       to the calling object which is the object that is passed around in event handlers.
 *
 *       Maybe we should apply all computed values/timestamps to the original object first,
 *       and then create the new object from that and just copy in the values that are supposed
 *       to be there?
 */
async function parseSaveObj(col, obj, opts) {
    const def = col.def;
    const timestamps = def.timestamps && (!opts || opts.timestamps !== false);
    if (timestamps) {
        const now = new Date();
        // Don't overwrite createdAt in case we are coming from an upsert update.
        obj.createdAt = obj.createdAt || now;
        obj.updatedAt = now;
    }
    const fields = await col.fieldsFor({ match: obj, static: true }), insertObj = new col();
    // this handles the case where they set "timestamps: true" but did not define the updatedAt/createdAt fields manually
    if (timestamps) {
        insertObj.createdAt = obj.createdAt;
        insertObj.updatedAt = obj.updatedAt;
    }
    _.each(col.denormal, function (field, name) {
        // we only need to copy it if it is a top-level level, nested values will be copied below
        if (name.indexOf('.') < 1) {
            name = path_1.default.populateNameFor(name, true);
            insertObj[name] = obj[name];
        }
    });
    _.each(fields, function (field, name) {
        const fieldDef = field.def;
        if (fieldDef.db !== false && !(fieldDef.get || fieldDef.getServer)) {
            const v = obj[name];
            if (v === undefined) {
                const df = _.result(fieldDef, 'defaultValue');
                if (df !== undefined) {
                    insertObj[name] = df;
                }
            }
            else {
                insertObj[name] = v;
            }
        }
    });
    const pkFieldName = def.primaryKey.field;
    if (insertObj[pkFieldName] === undefined) {
        const type = fields[pkFieldName].type;
        insertObj[pkFieldName] = type.generatePrimaryKeyVal();
    }
    if (def.primaryKey.defaultMatchIdOnInsert && insertObj._id === undefined) {
        insertObj._id = insertObj[def.primaryKey.field];
    }
    if (def.historical && obj._history) {
        insertObj._history = obj._history;
    }
    return insertObj;
}
exports.parseSaveObj = parseSaveObj;
function parseProjection(col, obj) {
    const def = col.def, projection = Object.assign({}, obj);
    // if given a projection like:   { foo: 1, 'foo.bar': 1 }
    // mongodb will only include foo.bar whereas we want to include all of foo.
    // So remove any projected paths here that subpaths of something else
    for (const name in projection) {
        if (projection.hasOwnProperty(name) && projection[name]) {
            for (let i = 0, ilen = name.length; i < ilen; i++) {
                if (name[i] === '.' && projection[name.substring(0, i)]) {
                    delete projection[name];
                    break;
                }
            }
        }
    }
    if (projection[def.primaryKey.field] === undefined &&
        // if an exclusion is present, don't add an inclusion
        _.find(projection, v => !v) === undefined) {
        projection[def.primaryKey.field] = 1;
    }
    // TODO: consider default removing _id if not the primary
    return projection;
}
exports.parseProjection = parseProjection;
function evaluateClient(client, key, doc, value, opts, proj) {
    if (_.isFunction(client)) {
        client = client.call(doc, value, opts, proj);
        // non-truthy from a function means no
        if (!client) {
            client = false;
        }
    }
    switch (client) {
        case false:
        case 0:
        case 'never':
            return false;
        case 'conditional':
            if (!proj || !proj[key]) {
                return false;
            }
        // fall through
        case true:
        case 1:
        case 'default':
        case undefined:
            return true;
        default:
            throw new Error(`Invalid client value: ${value}\n\nMust be one of 'never', 'conditional', 'default', true (or 1), false (or 0), or undefined`);
    }
}
exports.evaluateClient = evaluateClient;
function toClient(col, doc, opts) {
    if (!doc)
        return doc;
    if (Array.isArray(doc))
        return doc.map(doc2 => toClient(col, doc2, opts));
    if (doc.$access)
        doc.$redact();
    if (col)
        opts = processOptions(col, opts);
    const proj = projection_1.extractProjection(opts);
    // fields is only for top-level objects, we do not want to recursively pass it down into embedded documents
    const dOpts = proj ? _.omit(opts, 'fields', 'projection', 'project') : opts;
    const obj = {}, { populate } = opts;
    // note we are only looking at the top-level properties so call flattenProjection() before calling toClient()
    function projected(key) {
        var _a, _b, _c;
        if (key.endsWith('$')) {
            if (!populate)
                return false;
            let baseKey = key.substring(0, key.length - 1);
            if (populate[baseKey] || ((_a = populate.includes) === null || _a === void 0 ? void 0 : _a.call(populate, baseKey)))
                return true;
            // TODO:  this logic is hackish
            baseKey += '._';
            if (populate[baseKey] || ((_b = populate.includes) === null || _b === void 0 ? void 0 : _b.call(populate, baseKey)))
                return true;
            return false;
        }
        else {
            if (!proj)
                return key !== '_history' && key !== '$options';
            if (key === '_id')
                return proj[key] || proj[key] === undefined;
            const v = proj[key];
            if (v !== undefined)
                return true;
            if (populate) {
                const unpopulateName = key + 'Id';
                return populate[unpopulateName] || ((_c = populate.includes) === null || _c === void 0 ? void 0 : _c.call(populate, unpopulateName));
            }
        }
    }
    const fields = col === null || col === void 0 ? void 0 : col.fields;
    for (const k in doc) {
        if (!doc.hasOwnProperty(k))
            continue;
        let v = doc[k];
        let field;
        if (!projected(k))
            continue;
        if (fields && (field = fields[k])) {
            v = field.type.toClient(field, v, doc, opts, proj);
            if (v !== undefined) {
                obj[k] = v;
            }
        }
        else if (!v) {
            obj[k] = v;
        }
        else if (v.$toClient) {
            obj[k] = v.$toClient(dOpts);
        }
        else if (Array.isArray(v)) {
            // TODO:  we can figure out the type of k using metadata to make this work for the case when
            //        we have an array of pojos instead of instances
            obj[k] = toClient(null, v, dOpts);
        }
        else if (mongodb_1.ObjectId.isValid(v.toString())) {
            obj[k] = v.toString();
        }
        else {
            // TODO:  right now we're sending down everything we don't have metadata for ...
            //        for example, populated values ... we probably need a more comprehensive solution here, not sure
            //        what it would be yet
            obj[k] = v;
        }
    }
    if (doc.$access)
        obj.$access = doc.$access;
    // send down computed fields we can't compute on the client ...
    // maybe move everything into this so we only send down what we know about ... can also calculate populated names to send
    for (const name in fields) {
        if (!fields.hasOwnProperty(name))
            continue;
        const field = fields[name];
        const fieldDef = field.def;
        if ((fieldDef.get || fieldDef.getClient) && !opts.plain)
            continue;
        const getFn = fieldDef.getServer;
        if (!getFn)
            continue;
        if (!evaluateClient(fieldDef.client, name, doc, 
        // we don't pass in the value in this case because in order to do that we'd need to calculate it which might be costly
        undefined, opts, proj) ||
            !projected(name)) {
            continue;
        }
        let value;
        if (name in doc) {
            // Property will have been set on CollectionInstances
            value = doc[name];
        }
        else if (_.isFunction(getFn)) {
            // Have to set manually for POJO
            value = getFn.call(doc);
        }
        else {
            throw new Error('Incorrect computed value definition: ' + name);
        }
        obj[name] = value;
    }
    const toClientFn = col && col.def && col.def.toClient;
    if (toClientFn) {
        toClientFn.call(obj, opts);
    }
    const postFn = opts.post;
    if (postFn) {
        postFn.call(obj, opts);
    }
    return obj;
}
exports.toClient = toClient;
//# sourceMappingURL=common.js.map